; Copyright 2019-2021 Sean Porter (Glitchfinder)
;
; Permission is hereby granted, free of charge, to any person obtaining a copy
; of this software and associated documentation files (the "Software"), to deal
; in the Software without restriction, including without limitation the rights
; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
; copies of the Software, and to permit persons to whom the Software is
; furnished to do so, subject to the following conditions:
;
; The above copyright notice and this permission notice shall be included in all
; copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
; SOFTWARE.
;------------------------------------------------------------------------------/

Scriptname zYAWA_Manager_INST Extends Quest

;------------------------------------------------------------------------------\
; Data tracking objects
;------------------------------------------------------------------------------/

;CHECKLIST FOR CHANGING FACTIONS:
;1.Change TFS (recruit quest) stage.
;2.Change HSH (Kickout quest) stage (Ignore if non-betrayal, see 4.)
;3.If using Restore Mechanist requirements on line 284 and 285, set 66 and 67 to False. If replacing Eyebots and Robots, ignore this.
;4.If a non-betrayal faction (IE Mechanist Robots), comment out lines 277 and 437 ( CheckAlliance() )

zYAWA_Settlement_INST[] Property Settlements  Auto Hidden ; All tracked settlements
ObjectReference[]  Property BSettlements Auto Hidden ; Bethesda settlements
ObjectReference[]  Property Regions      Auto ; Regions being tracked
ObjectReference    Property DataPoint    Auto ; Data tracking spawn point
FormList           Property Doors        Auto ; Settlement entrances to move
FormList           Property Centers      Auto ; Settlements to move
Int[]              Property Densities    Auto Hidden ; Setting to value storage
Int[]              Property Products     Auto Hidden ; Setting to value storage
Form               Property Settlement   Auto ; Settlement tracking object
Form               Property Patrol       Auto ; Patrol tracking object
Form               Property PatrolShift  Auto ; Patrol shift tracking object

;------------------------------------------------------------------------------\
; Patrol package keywords
;------------------------------------------------------------------------------/
Keyword Property Follow1     Auto ; The first patrol member
Keyword Property Follow2     Auto ; The second patrol member
Keyword Property Follow3     Auto ; The third patrol member
Keyword Property Follow4     Auto ; The fourth patrol member
Keyword Property Follow5     Auto ; The fifth patrol member
Keyword Property DestStart   Auto ; The first patrol endpoint
Keyword Property DestEnd     Auto ; The second patrol endpoint
Keyword Property DoorStart   Auto ; The first patrol midpoint
Keyword Property DoorEnd     Auto ; The second patrol midpoint

;------------------------------------------------------------------------------\
; Mod Settings
;------------------------------------------------------------------------------/
Int  Property Mortality         = 0     Auto Hidden ; Are settlers mortal?
Int  Property OldMortality      = 0     Auto Hidden ; Setting cache
Bool Property UpdateMortality   = False Auto Hidden ; Do patrols need to update?
Bool Property Betrayal          = True  Auto Hidden ; Can patrols go hostile?
Bool Property Robots            = True Auto Hidden ; Can patrols have robots?
Bool Property Eyebots           = True Auto Hidden ; Can patrols have eyebots?
Bool Property UseSettlerBase    = True  Auto Hidden ; Use settlers or minutemen?
Bool Property OldUseSettlerBase = True  Auto Hidden ; Setting cache
Bool Property UpdateSettlerBase = False Auto Hidden ; Do patrols need to update?
Bool Property EyebotRadio       = False  Auto Hidden ; Do eyebots play the radio?
Bool Property MoveHomePlate     = False Auto Hidden ; Move home plate outside?
Bool Property OldMoveHomePlate  = False Auto Hidden ; Setting cache
Bool Property MoveMechanist     = False Auto Hidden ; Move the lair's door?
Bool Property OldMoveMechanist  = False Auto Hidden ; Setting cache
Bool Property MoveVault88       = False Auto Hidden ; Move Vault 88's door?
Bool Property OldMoveVault88    = False Auto Hidden ; Setting cache
Bool Property MoveSpectacle     = False Auto Hidden ; Move spectacle island?
Bool Property OldMoveSpectacle  = False Auto Hidden ; Setting cache

;------------------------------------------------------------------------------\
; DLC Data: Nuka World
;------------------------------------------------------------------------------/
Int            Property NukaCheck  Auto Hidden ; Countdown for DLC check
Bool           Property NukaWorld  Auto Hidden ; Is Nuka World installed?
Quest          Property HSHQuest   Auto ; Home Sweet Home quest
Quest          Property OSQuest    Auto Hidden ; Open Season quest
WorkshopScript Property NukaIgnore Auto Hidden ; Broken Nuka World settlement
Keyword        Property Raider     Auto Hidden ; Raider-owned workshop keyword
;------------------------------------------------------------------------------\
; DLC Data: Vault-Tec Workshop
;------------------------------------------------------------------------------/
Int  Property VaultCheck Auto Hidden ; Countdown for DLC check
Bool Property VaultTec   Auto Hidden ; Is Vault-Tec Workshop installed?
;------------------------------------------------------------------------------\
; DLC Data: Far Harbor
;------------------------------------------------------------------------------/
Int  Property HarborCheck Auto Hidden ; Countdown for DLC check
Bool Property FarHarbor   Auto Hidden ; Is Far Harbor installed?
;------------------------------------------------------------------------------\
; DLC Data: Automatron
;------------------------------------------------------------------------------/
Int   Property AutoCheck  Auto Hidden ; Countdown for DLC check
Bool  Property Automatron Auto Hidden ; Is Automatron installed?
Quest Property ROQuest    Auto Hidden ; Restoring Order quest
Quest Property MMQuest    Auto Hidden ; Mechanical Menace quest

;------------------------------------------------------------------------------\
; Worldspaces
;------------------------------------------------------------------------------/
WorldSpace Property Commonwealth Auto        ; The Commonwealth worldspace
WorldSpace Property NukaWorldWS  Auto Hidden ; Nuka World worldspace
WorldSpace Property TheIslandWS  Auto Hidden ; The Island worldspace

;------------------------------------------------------------------------------\
; Settlement Relocation Data
;------------------------------------------------------------------------------/
Cell            Property HomePlate         Auto        ; Home Plate's cell
ObjectReference Property HomePlateDoor     Auto        ; The speaker by the gate
Cell            Property SpectacleWorkshop Auto        ; Spectacle Island's cell
ObjectReference Property SpectacleDoor     Auto        ; A marker onshore
Cell            Property Vault88Workshop   Auto Hidden ; Vault 88's cell
ObjectReference Property Vault88Door1      Auto Hidden ; The default door
ObjectReference Property Vault88Door2      Auto Hidden ; The shop's door
Cell            Property MechanistLair     Auto Hidden ; The Mechanist's cell
ObjectReference Property MechanistDoor1    Auto Hidden ; The default door
ObjectReference Property MechanistDoor2    Auto Hidden ; The shop's door

;------------------------------------------------------------------------------\
; Rate limiting data
;------------------------------------------------------------------------------/
Int Property UpdatesPerTick  = 10 Auto Hidden ; How many patrols per tick?
Int Property UpdatedThisTick = 0  Auto Hidden ; How many patrols this tick?
Int Property SkippedThisTick = 0  Auto Hidden ; Where do we start next tick?
Int Property IndexThisTick   = 0  Auto Hidden ; Where are we this tick?
Int Property IndexLastTick   = 0  Auto Hidden ; Where did we start last tick?

;------------------------------------------------------------------------------\
; Actors to spawn
;------------------------------------------------------------------------------/
ActorBase Property PatrollerS Auto ; The settler-based patroller
ActorBase Property PatrollerM Auto ; The minutemen-based patroller
ActorBase Property Eyebot     Auto ; An eyebot for patrols
ActorBase Property Robot      Auto ; A robot for patrols

;------------------------------------------------------------------------------\
; General Data
;------------------------------------------------------------------------------/
OutputModel Property RadioModel  Auto Const  ; The eyebot radio model
Faction     Property Minutemen   Auto        ; A custom minutemen faction
Keyword     Property Center      Auto Hidden ; Workshop location keyword
Int         Property Food        Auto Hidden ; Workshop food keyword
Int         Property Water       Auto Hidden ; Workshop water keyword
Int         Property Safety      Auto Hidden ; Workshop defense keyword
Bool        Property Initialized Auto Hidden ; Has this mod been initialized?
Quest       Property TFSQuest    Auto        ; The First Steps quest
; The Bethesda-made workshop handling script
WorkshopParentScript Property WPS Auto Const Mandatory

;------------------------------------------------------------------------------\
; When this quest is created
;------------------------------------------------------------------------------/
Event OnInit()
	; If the game is not in menu mode, start the mod's timer loop
	If (!Utility.IsInMenuMode())
		StartTimerGameTime(0.033)
	; If the game is in menu mode, try again in one minute
	Else
		StartTimer(60)
	EndIf
EndEvent

;------------------------------------------------------------------------------\
; When the real-time timer runs out
;------------------------------------------------------------------------------/
Event OnTimer(Int aiTimerID)
	; If the game is not in menu mode, start the mod's timer loop
	If (!Utility.IsInMenuMode())
		StartTimerGameTime(0.033)
	; If the game is in menu mode, try again in one minute
	Else
		StartTimer(60)
	EndIf
EndEvent

;------------------------------------------------------------------------------\
; Initialize this mod
;------------------------------------------------------------------------------/
Function Initialize()
	; Do nothing if the mod is already set up
	If (Initialized)
		Return
	EndIf

	; Ensure this only runs once
	Initialized = True

	; Set up the option to data conversion tables
	Densities = New Int[5]
	Products  = New Int[5]
	Densities[0] = 0
	Densities[1] = 26624
	Densities[2] = 35498
	Densities[3] = 53248
	Densities[4] = -1
	Products[0]  = 0
	Products[1]  = 10
	Products[2]  = 40
	Products[3]  = 80
	Products[4]  = 200

	; Set up this mod's data storage
	Settlements  = New zYAWA_Settlement_INST[128]
	BSettlements = New ObjectReference[128]

	; Grab the bethesda-made workshop keywords
	Center = WPS.WorkshopLinkCenter
	Raider = WPS.WorkshopType02
	Food   = WPS.WorkshopRatingFood
	Water  = WPS.WorkshopRatingWater
	Safety = WPS.WorkshopRatingSafety

	; Check the player's installed DLC
	CheckDLC_NW()
	CheckDLC_FH()
	CheckDLC_A()
	CheckDLC_VT()

	; Move home plate
	MoveDoorC(HomePlate, HomePlateDoor)
EndFunction

;------------------------------------------------------------------------------\
; Update this mod's data
;------------------------------------------------------------------------------/
Event OnTimerGameTime(Int aiTimerID)
	; If the player has not become the heir of the Institute
	If (TFSQuest.GetStage() != 350)
		; Stop procseesing until the next check
		StartTimerGameTime(0.033)
		Return
	EndIf

	; Make sure this mod is set up
	Initialize()

	; If the rate-limit is on and the full list has been checked
	If (IndexLastTick == SkippedThisTick)
		; Reset the starting point
		SkippedThisTick = 0
	EndIf

	; Reset the rate-limiting data
	UpdatedThisTick = 0
	IndexThisTick   = 0
	IndexLastTick   = SkippedThisTick

	; If the mortality setting has changed
	If (Mortality != OldMortality)
		; Cache the setting and mark patrols as needing to update
		UpdateMortality = True
		OldMortality = Mortality
		UpdateMortality()
	EndIf

	; If the ActorBase setting has changed
	If (UseSettlerBase != OldUseSettlerBase)
		; Cache the setting and mark patrols as needing to update
		UpdateSettlerBase = True
		OldUseSettlerBase = UseSettlerBase
		UpdateActorBase()
	EndIf
	
	; If Nuka World was not installed, check to see if it is now
	If (NukaWorld != True)
		CheckDLC_NW()
	; If Nuka World is installed, update the faction hostility status
	Endif
		CheckAlliance()
;	EndIf

	; If Automatron was not installed, check to see if it is now
	If (Automatron != True)
		CheckDLC_A()
	; If Automatron is installed, check the robot quest statuses
	;Else
	;	CheckRobots()
	; We want "robots" (AKA Paladins) right out of the box, so no robot check.
	EndIf

	; If Far Harbor was not installed, check to see if it is now
	If (FarHarbor != True)
		CheckDLC_FH()
	EndIf

	; If Vault-Tec Workshop was not installed, check to see if it is now
	If (VaultTec != True)
		CheckDLC_VT()
	EndIf

	; Update the regional settings
	CheckRegions()
	; Check to see if any settlements have been moved
	CheckMoveOptions()
	; Update the workshops and their patrols
	CheckWorkshops()
	; Reset the regions' update flags
	ResetRegions()

	; Reset the general options' update flags
	UpdateMortality   = False
	UpdateSettlerBase = False

	; Set the next update
	StartTimerGameTime(0.033)
EndEvent

;------------------------------------------------------------------------------\
; Check regional options for updates
;------------------------------------------------------------------------------/
Function CheckRegions()
	; Start with the first region
	Int i = 0

	; Check every region in a row
	While (i < 8)
		; Get the current region
		zYAWA_Region_INST r = (Regions[i] As zYAWA_Region_INST)
		; Update the region
		r.Update()
		; If this region has updated
		If (r.Updated)
			; Update the region's patrols
			UpdatePatrolSettings(r)
		EndIf
		; Move on to the next region
		i += 1
	EndWhile
EndFunction

;------------------------------------------------------------------------------\
; Reset the regional option update flags
;------------------------------------------------------------------------------/
Function ResetRegions()
	; Start with the first region
	Int i = 0

	; Check every region in a row
	While (i < 8)
		; Reset the update flag and move on to the next
		(Regions[i] As zYAWA_Region_INST).Updated = False
		i += 1
	EndWhile
EndFunction

;------------------------------------------------------------------------------\
; Check to see if a player has opted to move or un-move a settlement
;------------------------------------------------------------------------------/
Function CheckMoveOptions()
	; If the Home Plate option has changed
	If (MoveHomePlate != OldMoveHomePlate)
		; Cache the option's new status
		OldMoveHomePlate = MoveHomePlate

		; If Home Plate is being moved, move it
		If (MoveHomePlate)
			MoveCenterC(HomePlate, HomePlateDoor)
		; If Home Plate is being unmoved, restore it
		Else
			UnmoveCenterC(HomePlate)
		EndIf
	EndIf

	; If the Mechanist's Lair option has changed
	If (Automatron && MoveMechanist != OldMoveMechanist)
		; Cache the option's new status
		OldMoveMechanist = MoveMechanist

		; If The Mechanist's Lair is being moved, move it
		If (MoveMechanist)
			MoveDoorC(MechanistLair, MechanistDoor2)
		; If The Mechanist's Lair is being unmoved, restore it
		Else
			MoveDoorC(MechanistLair, MechanistDoor1)
		EndIf
	EndIf

	; If the Vault 88 option has changed
	If (VaultTec && MoveVault88 != OldMoveVault88)
		; Cache the option's new status
		OldMoveVault88 = MoveVault88

		; If Vault 88 is being moved, move it
		If (MoveVault88)
			MoveDoorC(Vault88Workshop, Vault88Door2)
		; If Vault 88 is being unmoved, restore it
		Else
			MoveDoorC(Vault88Workshop, Vault88Door1)
		EndIf
	EndIf

	; If the Spectacle Island option has changed
	If (MoveSpectacle != OldMoveSpectacle)
		; Cache the option's new status
		OldMoveSpectacle = MoveSpectacle

		; If Spectacle Island is being moved, move it
		If (MoveSpectacle)
			MoveDoorC(SpectacleWorkshop, SpectacleDoor)
			MoveCenterC(SpectacleWorkshop, SpectacleDoor)
		; If Spectacle Island is being unmoved, restore it
		Else
			UnmoveDoorC(SpectacleWorkshop)
			UnmoveCenterC(SpectacleWorkshop)
		EndIf
	EndIf
EndFunction

;------------------------------------------------------------------------------\
; Check to see if Nuka World is installed
;------------------------------------------------------------------------------/
Function CheckDLC_NW()
    ; Make sure this check doesn't run constantly
    If (NukaCheck > 0)
        NukaCheck -= 1
        Return
    EndIf

    ; Set the DLC's filename
    String n = "DLCNukaWorld.esm"

    ; Try to load the Home Sweet Home quest from file
    OSQuest     = Game.GetFormFromFile(0x00027704, n) as Quest
    ; Set the Nuka World installation status based on whether it loaded
    NukaWorld = (OSQuest != None)

    ; If Nuka World is not installed, prevent another check for a while
    If (NukaWorld != True)
        NukaCheck = 50
    ; If Nuka World is installed
    Else
        ; Load the rest of the desired data from file
        NukaWorldWS = Game.GetFormFromFile(0x0000290F, n) as WorldSpace
        NukaIgnore  = Game.GetFormFromFile(0x00047DFB, n) as WorkshopScript

        ; Check to see if the minutemen should be hostile to the player
        CheckAlliance()
    EndIf
EndFunction

;------------------------------------------------------------------------------\
; Check to see if Far Harbor is installed
;------------------------------------------------------------------------------/
Function CheckDLC_FH()
	; Make sure this check doesn't run constantly
	If (HarborCheck > 0)
		HarborCheck -= 1
		Return
	EndIf

	; Try to load Far Harbor's worldspace from file
	TheIslandWS = Game.GetFormFromFile(0x00000B0F, "DLCCoast.esm") as WorldSpace
	; Set the Far Harbor installation status based on whether it loaded
	FarHarbor = (TheIslandWS != None)

	; If Far Harbor is not installed, prevent another check for a while
	If (FarHarbor != True)
		HarborCheck = 50
	EndIf
EndFunction

;------------------------------------------------------------------------------\
; Check to see if Automatron is installed
;------------------------------------------------------------------------------/
Function CheckDLC_A()
	; Make sure this check doesn't run constantly
	If (AutoCheck > 0)
		AutoCheck -= 1
		Return
	EndIf

	; Set the DLC's filename
	String n = "DLCRobot.esm"

	; Try to load the Mechanical Menace quest from file
	MMQuest = Game.GetFormFromFile(0x00000806, n) as Quest
	; Set the Automatron installation status based on whether it loaded
	Automatron = (MMQuest != None)

	; If Automatron is not installed, prevent another check for a while
	If (Automatron != True)
		AutoCheck = 50
	; If Automatron is installed
	Else
		; Load the rest of the desired data from file
		ROQuest        = Game.GetFormFromFile(0x000010F5, n) as Quest
		MechanistLair  = Game.GetFormFromFile(0x000008A3, n) as Cell
		MechanistDoor1 = Game.GetFormFromFile(0x0000356E, n) as ObjectReference
		MechanistDoor2 = Game.GetFormFromFile(0x00004317, n) as ObjectReference

		; Set the entrance location for The Mechanist's Lair
		MoveDoorC(MechanistLair, MechanistDoor1)
		; Check the status of the robot quests and their options
		;CheckRobots()
		; Again, we want 'robots' (PA Soldiers) right out of the box, so no robot check necessary.
	EndIf
EndFunction

;------------------------------------------------------------------------------\
; Check to see if Vault-Tec Workshop is installed
;------------------------------------------------------------------------------/
Function CheckDLC_VT()
	; Make sure this check doesn't run constantly
	If (VaultCheck > 0)
		VaultCheck -= 1
		Return
	EndIf

	; Set the DLC's filename
	String n = "DLCworkshop03.esm"

	; Attempt to load the Vault 88 cell from file
	Vault88Workshop = Game.GetFormFromFile(0x00000F31, n) as Cell
	; Set the Vault-Tec Workshop installation status based on whether it loaded
	VaultTec = (Vault88Workshop != None)

	; If Vault-Tec Workshop is not installed, prevent another check for a while
	If (VaultTec != True)
		VaultCheck = 50
	; If Vault-Tec Workshop is installed
	Else
		; Load the rest of the desired data from file
		Vault88Door1 = Game.GetFormFromFile(0x00003CC2, n) as ObjectReference
		Vault88Door2 = Game.GetFormFromFile(0x00003CC5, n) as ObjectReference

		; Set the entrance location for Vault 88
		MoveDoorC(Vault88Workshop, Vault88Door1)
	EndIf
EndFunction

;------------------------------------------------------------------------------\
; Check to see the status of relevant Automatron quests
;------------------------------------------------------------------------------/
Function CheckRobots()
	; If Mechanical Menace is complete, the player has access to eyebots
	If (MMQuest.GetStage() >= 3000)
		Eyebots = True
	EndIf

	; If Restoring Order is complete, the player has access to robots
	If (ROQuest.GetStage() >= 1100)
		Robots = True
	EndIf
EndFunction

;------------------------------------------------------------------------------\
; Check InstKickOut to see if the Institute are hostile
;------------------------------------------------------------------------------/
Function CheckAlliance()
	; If the betrayal option is disabled
	If (!Betrayal)
		; If the minutemen are hostile, reset their mortality
		If (Minutemen.IsPlayerEnemy())
			UpdateMortality = True
		EndIf

		; Set the minutemen to friendly and exit processing
		Minutemen.SetPlayerEnemy(False)
		Return
	EndIf

	; If Home Sweet Home is incomplete, they are not hostile
	If (HSHQuest.GetStage() < 200)
		; Set the minutemen to friendly and exit processing
		Minutemen.SetPlayerEnemy(False)
		Return
	EndIf

	; If Open Season is complete
	;If (OSQuest.GetStage() >= 255)
	;	; If the minutemen are hostile, reset their mortality
	;	If (Minutemen.IsPlayerEnemy() != False)
	;		UpdateMortality = True
	;	EndIf
	;
	;	; Set the minutemen to friendly and exit processing
	;	Minutemen.SetPlayerEnemy(False)
	;	Return
	;EndIf
	; No redemption for the Institute.

	; If Home Sweet Home is complete and Open Season is not, they are hostile
	If (Minutemen.IsPlayerEnemy() != True)
		; Ensure the minutemen are mortal
		UpdateMortality = True
	EndIf

	; Ensure the minutemen are hostile
	Minutemen.SetPlayerEnemy(True)
EndFunction

;------------------------------------------------------------------------------\
; Check and update all workshops
;------------------------------------------------------------------------------/
Function CheckWorkshops()
	; Get the number of workshops in the game
	Int l = WPS.Workshops.Length
	; Start with the first workshop
	Int i = -1

	; Go through every workshop in the list, up to 128
	While i < l - 1
		; Raise the index by one
		i += 1

		; Get the current workshop to check
		WorkshopScript w = WPS.Workshops[i] As WorkshopScript

		; If the workshop has not already been handled
		If (w != None && BSettlements.Find(w As ObjectReference) < 0)
			; If this is a valid workshop
			If (w.MyLocation != None && (!NukaWorld || w != NukaIgnore))
				; Add, track, and initialize this settlement
				AddWorkshop(w, i)
			EndIf
		EndIf

		; If this settlement exists
		If (Settlements[i] != None)
			; Update it
			Settlements[i].Update(Self, i)
		EndIf
	EndWhile
EndFunction

;------------------------------------------------------------------------------\
; Check and update a patrol's settings
;------------------------------------------------------------------------------/
Function UpdatePatrolSettings(zYAWA_Region_INST r)
	; Get the number of workshops in the game
	Int l = WPS.Workshops.Length
	; Start with the first workshop
	Int i = -1

	; Go through every workshop in the list, up to 128
	While i < l - 1
		; Raise the index by one
		i += 1

		; If this settlement exists
		If (Settlements[i] != None)
			; Update it
			Settlements[i].Settings(Self, r)
		EndIf
	EndWhile
EndFunction

;------------------------------------------------------------------------------\
; Check and update a patrol's mortality
;------------------------------------------------------------------------------/
Function UpdateMortality()
	; Get the number of workshops in the game
	Int l = WPS.Workshops.Length
	; Start with the first workshop
	Int i = -1

	; Go through every workshop in the list, up to 128
	While i < l - 1
		; Raise the index by one
		i += 1

		; If this settlement exists
		If (Settlements[i] != None)
			; Update it
			Settlements[i].Mortality(Self)
		EndIf
	EndWhile
EndFunction

;------------------------------------------------------------------------------\
; Check and update a patrol's mortality
;------------------------------------------------------------------------------/
Function UpdateActorBase()
	; Get the number of workshops in the game
	Int l = WPS.Workshops.Length
	; Start with the first workshop
	Int i = -1

	; Go through every workshop in the list, up to 128
	While i < l - 1
		; Raise the index by one
		i += 1

		; If this settlement exists
		If (Settlements[i] != None)
			; Update it
			Settlements[i].BaseActor(Self)
		EndIf
	EndWhile
EndFunction

;------------------------------------------------------------------------------\
; Add a new workshop to be tracked
;------------------------------------------------------------------------------/
Function AddWorkshop(WorkshopScript w, Int i)
	; Add this workshop to the list of bethesda workshops
	BSettlements[i] = w
	; Add our own tracking data
	Settlements[i] = DataPoint.PlaceAtMe(Settlement) As zYAWA_Settlement_INST
	; Initialize the workshop's data
	Settlements[i].Initialize(Self, w, i)
EndFunction

;------------------------------------------------------------------------------\
; Sort a settlement by region
;------------------------------------------------------------------------------/
Int Function SortSettlement(ObjectReference ref)
	; If the settlement doesn't exist, it is in the Unknown region
	If (ref == None)
		Return 8
	EndIf

	; If the settlement is not in the commonwealth, use the unknown subsort
	If (ref.GetWorldSpace() != Commonwealth)
		Return SubsortSettlement(ref.GetWorldSpace())
	EndIf

	; Get the settlement's location
	Float x = ref.X
	Float y = ref.Y

	; The default result the northwest commonwealth
	Int result = 0
	Float minDist = Math.sqrt(Math.pow(x + 60740, 2) + Math.pow(y - 73823, 2))

	; Check to see if it is in the central divide
	Float dist = Math.sqrt(Math.pow(x + 3086, 2) + Math.pow(y - 67594, 2))
	If (dist < minDist)
		minDist = dist
		result = 1
	EndIf

	; Check to see if it is in the northeast commonwealth
	dist = Math.sqrt(Math.pow(x - 77892, 2) + Math.pow(y - 15089, 2))
	If (dist < minDist)
		minDist = dist
		result = 2
	EndIf

	; Check to see if it is in the western reach
	dist = Math.sqrt(Math.pow(x + 70606, 2) + Math.pow(y + 25075, 2))
	If (dist < minDist)
		minDist = dist
		result = 3
	EndIf

	; Check to see if it is downtown
	dist = Math.sqrt(Math.pow(x - 13310, 2) + Math.pow(y + 18905, 2))
	If (dist < minDist)
		minDist = dist
		result = 4
	EndIf

	; Check to see if it is in the great southern expanse
	dist = Math.sqrt(Math.pow(x - 62454, 2) + Math.pow(y + 82489, 2))
	If (dist < minDist)
		minDist = dist
		result = 5
	EndIf

	; Return the settlement's region in the commonwealth
	Return result
EndFunction

;------------------------------------------------------------------------------\
; Sort a non-Commonwealth settlement by region
;------------------------------------------------------------------------------/
Int Function SubsortSettlement(Worldspace w)
	; If the settlement is in nuka world, it is in the nuka world region
	If (NukaWorld && w == NukaWorldWS)
		Return 6
	EndIf

	; If the settlement is on the island, it is in The Island region
	If (FarHarbor && w == TheIslandWS)
		Return 7
	EndIf

	; The settlement's region is unknown
	Return 8
EndFunction

;------------------------------------------------------------------------------\
; Move a settlement entrance
;------------------------------------------------------------------------------/
Function MoveDoorC(Cell WorkshopCell, ObjectReference NewDoor)
	; Set up the patrol change
	zYAWA_PatrolShift_INST temp = GetPatrolShift()
	temp.ParentCell = WorkshopCell
	temp.NewPoint = NewDoor
	; Move the patrol
	Doors.AddForm(temp)
EndFunction

;------------------------------------------------------------------------------\
; Move a settlement entrance
;------------------------------------------------------------------------------/
Function MoveDoorW(WorkshopScript Workshop, ObjectReference NewDoor)
	; Set up the patrol change
	zYAWA_PatrolShift_INST temp = GetPatrolShift()
	temp.Workshop = Workshop
	temp.NewPoint = NewDoor
	; Move the patrol
	Doors.AddForm(temp)
EndFunction

;------------------------------------------------------------------------------\
; Reset a settlement entrance
;------------------------------------------------------------------------------/
Function UnmoveDoorC(Cell WorkshopCell)
	; Set up the patrol change
	zYAWA_PatrolShift_INST temp = GetPatrolShift()
	temp.ParentCell = WorkshopCell
	temp.Undo = True
	; Move the patrol
	Doors.AddForm(temp)
EndFunction

;------------------------------------------------------------------------------\
; Reset a settlement entrance
;------------------------------------------------------------------------------/
Function UnmoveDoorW(WorkshopScript Workshop)
	; Set up the patrol change
	zYAWA_PatrolShift_INST temp = GetPatrolShift()
	temp.Workshop = Workshop
	temp.Undo = True
	; Move the patrol
	Doors.AddForm(temp)
EndFunction

;------------------------------------------------------------------------------\
; Move a settlement center
;------------------------------------------------------------------------------/
Function MoveCenterC(Cell WorkshopCell, ObjectReference NewCenter)
	; Set up the patrol change
	zYAWA_PatrolShift_INST temp = GetPatrolShift()
	temp.ParentCell = WorkshopCell
	temp.NewPoint = NewCenter
	; Move the patrol
	Centers.AddForm(temp)
EndFunction

;------------------------------------------------------------------------------\
; Move a settlement center
;------------------------------------------------------------------------------/
Function MoveCenterW(WorkshopScript Workshop, ObjectReference NewCenter)
	; Set up the patrol change
	zYAWA_PatrolShift_INST temp = GetPatrolShift()
	temp.Workshop = Workshop
	temp.NewPoint = NewCenter
	; Move the patrol
	Centers.AddForm(temp)
EndFunction

;------------------------------------------------------------------------------\
; Reset a settlement center
;------------------------------------------------------------------------------/
Function UnmoveCenterC(Cell WorkshopCell)
	; Set up the patrol change
	zYAWA_PatrolShift_INST temp = GetPatrolShift()
	temp.ParentCell = WorkshopCell
	temp.Undo = True
	; Move the patrol
	Centers.AddForm(temp)
EndFunction

;------------------------------------------------------------------------------\
; Reset a settlement center
;------------------------------------------------------------------------------/
Function UnmoveCenterW(WorkshopScript Workshop)
	; Set up the patrol change
	zYAWA_PatrolShift_INST temp = GetPatrolShift()
	temp.Workshop = Workshop
	temp.Undo = True
	; Move the patrol
	Centers.AddForm(temp)
EndFunction

;------------------------------------------------------------------------------\
; Get a patrol redirect object
;------------------------------------------------------------------------------/
zYAWA_PatrolShift_INST Function GetPatrolShift()
	Return (DataPoint.PlaceAtMe(PatrolShift, 1, true) As zYAWA_PatrolShift_INST)
EndFunction
